// parser.cc
//
//

#include <forint/forint.hh>

namespace FInt {

namespace FInt_internal {

const
char* str_type [] = { "char", "short", "int", "long", "long long",
		      "float", "double", "long double" };

const uint num_coor = 12;
const
char* cmd [] = { "coor", "x", "y", "z", "phi", "psi", "r", "rad",
		         "t", "u", "v", "rl", "self",
		 "vrbl",
		 "poin",
		 "numb", "M_E", "M_LOG2E", "M_LOG10E", "M_LN2", "M_LN10",
		         "M_PI", "M_PI_2", "M_PI_4", "M_1_PI", "M_2_PI",
		         "M_2_SQRTPI", "M_SQRT2", "M_SQRT1_2",
		 "func", "func1", "func2d", "func2p", "func2",
		         "sin", "cos", "tan", "asin", "acos", "atan",
		         "sinh", "cosh", "tanh", "asinh", "acosh", "atanh",
		         "exp", "log", "log10", "sqrt",
		         "expm1", "log1p", "cbrt",
		         "j0", "j1", "y0", "y1", "erf", "erfc", "lgamma",
		         "fabs", "abs", "labs", "ceil", "floor", "rint",
		         "!", "~", "@-", "@+",
		 "oper", "+", "-", "*", "/", "%",
		         "&", "|", "^", "<<", ">>",
		         "atan2", "fmod", "pow", "hypot",
		         "ldexp", "drem", "copysign", "jn", "yn", 
		 "cond", ">", ">=", "<", "<=", "==", "!=",
		 "logc", "&&", "||",
		 "gram", "if", "else", "while", "cont", "jump",
                         ";", ",", "(", ")",
		 "=",    "+=", "-=", "*=", "/=", "%=", "&=", "|=", "^=",
                         "<<=",">>=" };

uint base = 10;

template vec_t <USR_FUNC_0_PTR>;
template vec_t <USR_FUNC_1_PTR>;
template vec_t <USR_FUNC_2D_PTR>;
template vec_t <USR_FUNC_2P_PTR>;
template vec_t <USR_FUNC_2_PTR>;
template vec_t <action_t>;
template vec_t <poin_t>;
template vec_t <double>;
template vec_t <char*>;
template vec_t <uint>;
template vec_t <bool>;
template stack_t <action_t>;
template stack_t <double>;
template stack_t <void*>;
template stack_t <uint>;

}

// producing rules (S -- original expression, L -- left value):
//
//   0-th priority:
//     S -> L  = S ;
//     S -> L += S ;
//     S -> L -= S ;
//     S -> L *= S ;
//     S -> L /= S ;
//     S -> L %= S ;
//     S -> L &= S ;
//     S -> L |= S ;
//     S -> L ^= S ;
//     S -> L <<= S ;
//     S -> L >>= S ;
//
//   1-st priority:
//     S -> T1 ? T1 : T1 ;
//     S -> T1
//
//   2-nd priority:
//     T1 -> T2 || T1
//     T1 -> T2
//
//   3-rd priority:
//     T2 -> T3 && T2
//     T2 -> T3
//
//   4-th priority:
//     T3 -> T4 | T3
//     T3 -> T4
//
//   5-th priority:
//     T4 -> T5 ^ T4
//     T4 -> T5
//
//   6-th priority:
//     T5 -> T6 & T5
//     T5 -> T6
//
//   7-th priority:
//     T6 -> T7 == T6
//     T6 -> T7 != T6
//     T6 -> T7
//
//   8-th priority:
//     T7 -> T8 <  T7
//     T7 -> T8 <= T7
//     T7 -> T8 >  T7
//     T7 -> T8 >= T7
//     T7 -> T8
//
//   9-th priority:
//     T8 -> U << T8
//     T8 -> U >> T8
//     T8 -> U
//
//   10-th priority:
//     U -> V + U
//     U -> V - U
//     U -> V
//
//   11-th priority:
//     V -> W * V
//     V -> W / V
//     V -> W % V
//     V -> W
//
//   12-th priority:
//     W -> ( S )
//     W -> ! W
//     W -> - W
//     W -> + W
//     W -> <number>
//     W -> <external function> ()
//     W -> <external function> ( S )
//     W -> <external function> ( S , S )
//     W -> <external function> ( <external pointer> , <external pointer> )
//     W -> <external function> ( S , <external pointer> )
//     W -> <external variable>
//     W -> <external pointer>
//     W -> <external pointer> [ S ]...[ S ] (n-dimentional array)
//     W -> <built-in func> ( S )
//     W -> <built-in func> ( S , S )
//     W -> <built-in variable>
//     W -> <built-in constant>

// counter of dynamically linked functions
uint forint::dl_func_cntr = 0;


char* forint::
get_C_string(char* tmp_expr,
	     uint size)
{
  char* tmp = (char*)malloc(size+1*sizeof(char));
  strncpy(tmp,tmp_expr,size);
  tmp[size] = '\0';
  char* t = tmp;
  c.setStr(tmp_expr,strlen(tmp_expr));

  while (*t) {
    if (*t == '\n')
      *t = ' ';
    if (*t == '\t')
      *t = ' ';
    if (*t == ';')
      *t = ' ';
    ++t;
  }
  return tmp;
}


// 0-string -- string containing (semicolon ended) assignement statements
// or no one of them:
//   x += y /= z;
//   x + y
//   (r > 10) ? z : sin(phi)
int forint::
get_size_0_string(char* expr,
		  uint& size)
{
  if (!*expr) {
    size = 0;
    return 0;
  }
  char* tmp = expr;
  size = 0;
  while(*tmp) {
    if (*tmp == ';') {
      ++size;
      return 0;
    }
    if (*tmp == '(')
      if (pair(++tmp,strlen(tmp)-1,'(',')')) {
	error_code = 400;
	return error_code;
      }
    if (*tmp == '[')
      if (pair(++tmp,strlen(tmp)-1,'[',']')) {
	error_code = 401;
	return error_code;
      }
    size = ++tmp - expr;
  }
  return 0;
}

int forint::
parse(char* expr,
      parse_mode_t mode = NEW_FORM_MODE,
      bool bvar_access = true,
      bool ebvar_access = true,
      bool ctrl_access = true,
      bool clear = false)
{
  flag_of_built_in_var_access = bvar_access;
  flag_of_extra_built_in_var_access = ebvar_access;
  flag_of_ctrl_statement_access = ctrl_access;
  clearing_flag = clear;
  int rv = 0;
  uint size = 1;
  using FInt_internal::num_coor;
  char* tmp_expr = expr;
  uint r_action_s;
  uint s_action_s;
  uint number_s;
  uint offsets_s;
  error_code = 0;

  if (math_expression &&
      !strncmp(math_expression,expr,strlen(math_expression)) &&
      strlen(math_expression) == strlen(expr)) {
    return 0;
  }

  if (mode == NEW_FORM_MODE)
    clean();

  char* tmp_string = 0;
  uint offset = 0;
  uint m = 0;
  is_while = false;
  c.setStr(tmp_expr,strlen(tmp_expr));

  char* t = expr;
  while (*t) {
    if (*(t+1))
      if (*t == '+' && *(t+1) == '+' ||
	  *t == '-' && *(t+1) == '-') {
if (!locale) {
  c.err(t,2,"increment and decrement can't be used.\n");
  c.warn("Use assignement statements (");
  c.warn("+=",5);
  c.warn(", ");
  c.warn("-=",5);
  c.warn(") instead.\n");
} else {
  c.err(t,2,"       .\n");
  c.warn("   (");
  c.warn("+=",5);
  c.warn(", ");
  c.warn("-=",5);
  c.warn(")   .\n");
}
 c.flsh(); 
 return error_code = 399;
}
    ++t;
  }

// remember old values...
  r_action_s = r_action.size();
  s_action_s = s_action.size();
  number_s   = number.size();
  offsets_s = offsets.size();

// translation is made 0-string by 0-string...
  while (!(rv=get_size_0_string(tmp_expr,size)) && size) {
    tmp_string = get_C_string(tmp_expr,size);

// push_cast_pos() uses differences between its arguments and b_decomposed_string
    b_decomposed_string = tmp_string - offset;
    c.setStr(tmp_string,strlen(tmp_string));
    is_while = false;
    if ((rv=decompose(tmp_string, tmp_string+strlen(tmp_string), -1, m))) {

      free(tmp_string);
      break;
    }
    free(tmp_string);
    if (math_expression) {
      char* tmp = (char*)malloc((strlen(math_expression)+size+1)*
				sizeof(char));
      strcpy(tmp, math_expression);
      strncat(tmp, tmp_expr, size);
      tmp[strlen(math_expression)+size] = '\0';
      free(math_expression);
      math_expression = tmp;
    } else {
      math_expression = (char*)malloc(size+1);
      strncpy(math_expression, tmp_expr, size);
      math_expression[size] = '\0';
    }
    tmp_expr += size;
    offset += size;
  }
  if (!rv && is_while) {
if (!locale) {
  c.err(tmp_string,strlen(tmp_string),"the \'");
  c.err("while",5);
  c.err("\' statement can't end an expression.\n");
} else {
  c.err(tmp_string,strlen(tmp_string),"\'");
  c.err("while",5);
  c.err("\'     .\n");
}
 c.flsh(); 
 rv = error_code = 497;
}

  if (rv) {

// clean the structures...
    if (mode == NEW_FORM_MODE)
      clean();

// or remove additions from invalid expression...
    else {
      while (r_action_s > r_action.size())
	r_action.pop_back();
      while (s_action_s > s_action.size())
	s_action.pop_back();
      while (number_s > number.size())
	number.pop_back();
      while (offsets_s > offsets.size())
	offsets.pop();
    }
    return rv;
  }
  if (clearing_flag) {
    for (uint i=0; i<num_coor-1; i++)
      argument[i] = 0;
    argument[FI_RL] = 1;
  }
  error_code = 0;
  return rv;
}

// make DL formula, return zero on successful completing
int forint::
makeDLF(vector<char*>* dlf_opt=0)
{
  using namespace FInt_internal;
  int rv = 0;
  char* error = 0;
  error_code = 0;

  if (!dlfunc_lib) {
    if ((rv=mk_dl_func(math_expression,
		       &dlfunc_cc_name,
		       &dlfunc_so_name,
		       dlf_opt)))
      return rv;
    dlfunc_lib = dlopen (dlfunc_so_name, RTLD_LAZY);
    if (!dlfunc_lib) {
      fputs ("\nforint: dlopen() error\n", stderr);
      fputs (dlerror(), stderr);
      error_code = 1010;
      return error_code;
    }
/*
let we've compiled function "func(double)" by a g++ compiler, then "func__Fd"
symbol we find in obj code corresponding to the function.
let's consider the symbol. a part before "__" isn't interesting because it
is just the function's name, postfix "__Fd" which appeared in c++ obj code
because of function overloading in C++ consists of two part:
 "__" which is a delimiter of the function's name and argument part, and
 "Fd" which consists of "F" probably meaning "function" and "d" meaning a
      double argument.

examples of symbols for another signatures:
 "func__FPdPPPiPc" -- pointer to double, pointer to pointer to pointer to int,
      pointer to char,
 "func__FPt12basic_string3ZcZt18string_char_traits1ZcZt24__default_alloc_template2b1i0" -- a pointer to string,
 "func__Fv" -- void signature.

of course, symbols for pointers and array names have no postfices, because you
can't overload them.
*/
    dl_function = (double (*)(void)) dlsym(dlfunc_lib,
					     "forint_dlf__Q24FInt6forint");

    if ((error = dlerror()) != NULL)  {
      fputs(error, stderr);
      error_code = 1011;
      return error_code;
    }
    ++dl_func_cntr;
  }
  return 0;
}


char** forint::
getRProg(int& length)
{
  error_code = 0;
  using namespace FInt_internal;
  number_p      = number.getPtr();
  var_num_p     = var_num.getPtr();
  poin_num_p    = poin_num.getPtr();
  usr_function_0_num_p = usr_function_0_num.getPtr();
  usr_function_1_num_p = usr_function_1_num.getPtr();
  usr_function_2d_num_p = usr_function_2d_num.getPtr();
  usr_function_2p_num_p = usr_function_2p_num.getPtr();
  usr_function_2_num_p = usr_function_2_num.getPtr();
  action_t* prog = r_action.getPtr();
  if (!r_prog && r_action.size()) {
    r_prog = (char**)malloc(r_action.size() * sizeof(char*));
    for (uint i=0; i<r_action.size(); i++) {
      if (prog[i] == FI_UN_MINUS)
	r_prog[i] = strdup("unary-");
      else if (prog[i] == FI_NUMB) {
	char tmp[50];
	ostrstream out (tmp,50);
	out.precision(15);
	out << *number_p++ << '\0';
	r_prog[i] = strdup(tmp);
      }
      else if (prog[i] == FI_VRBL)
	r_prog[i] = strdup(var_name.getPtr()[*var_num_p++]);
      else if (prog[i] == FI_POIN)
	r_prog[i] = strdup(poin_name.getPtr()[*poin_num_p++]);
      else if (prog[i] == FI_FUNC)
	r_prog[i] = strdup(usr_function_0_name.getPtr()
			   [*usr_function_0_num_p++]);
      else if (prog[i] == FI_FUNC1)
	r_prog[i] = strdup(usr_function_1_name.getPtr()
			   [*usr_function_1_num_p++]);
      else if (prog[i] == FI_FUNC2D)
	r_prog[i] = strdup(usr_function_2d_name.getPtr()
			   [*usr_function_2d_num_p++]);
      else if (prog[i] == FI_FUNC2P)
	r_prog[i] = strdup(usr_function_2p_name.getPtr()
			   [*usr_function_2p_num_p++]);
      else if (prog[i] == FI_FUNC2)
	r_prog[i] = strdup(usr_function_2_name.getPtr()
			   [*usr_function_2_num_p++]);
      else if (prog[i] == FI_COOR)
	r_prog[i] = strdup("dummy");
      else
	r_prog[i] = strdup(FInt_internal::cmd[r_action.getPtr()[i]]);
    }
  }
  length = r_action.size();
  return r_prog;
}

char** forint::
getSProg(int& length)
{
  error_code = 0;
  using namespace FInt_internal;
  number_p      = number.getPtr();
  var_num_p     = var_num.getPtr();
  poin_num_p    = poin_num.getPtr();
  usr_function_0_num_p = usr_function_0_num.getPtr();
  usr_function_1_num_p = usr_function_1_num.getPtr();
  usr_function_2d_num_p= usr_function_2d_num.getPtr();
  usr_function_2p_num_p= usr_function_2p_num.getPtr();
  usr_function_2_num_p = usr_function_2_num.getPtr();
  action_t* prog = s_action.getPtr();
  if (!s_prog && s_action.size()) {
    s_prog = (char**)malloc(s_action.size() * sizeof(char*));
    for (uint i=0; i<s_action.size(); i++)
      if (prog[i] == FI_UN_MINUS)
	s_prog[i] = strdup("unary-");
      else if (prog[i] == FI_NUMB) {
	char tmp[50];
	ostrstream out (tmp,50);
	out.precision(15);
	out << *number_p++ << '\0';
	s_prog[i] = strdup(tmp);
      }
      else if (prog[i] == FI_VRBL)
	s_prog[i] = strdup(var_name.getPtr()[*var_num_p++]);
      else if (prog[i] == FI_POIN)
	s_prog[i] = strdup(poin_name.getPtr()[*poin_num_p++]);
      else if (prog[i] == FI_FUNC)
	s_prog[i] = strdup(usr_function_0_name.getPtr()
			   [*usr_function_0_num_p++]);
      else if (prog[i] == FI_FUNC1)
	s_prog[i] = strdup(usr_function_1_name.getPtr()
			   [*usr_function_1_num_p++]);
      else if (prog[i] == FI_FUNC2D)
	s_prog[i] = strdup(usr_function_2d_name.getPtr()
			   [*usr_function_2d_num_p++]);
      else if (prog[i] == FI_FUNC2P)
	s_prog[i] = strdup(usr_function_2p_name.getPtr()
			   [*usr_function_2p_num_p++]);
      else if (prog[i] == FI_FUNC2)
	s_prog[i] = strdup(usr_function_2_name.getPtr()
			   [*usr_function_2_num_p++]);
      else if (prog[i] == FI_COOR)
	s_prog[i] = strdup("dummy");
      else
	s_prog[i] = strdup(FInt_internal::cmd[s_action.getPtr()[i]]);
  }
  length = s_action.size();
  return s_prog;
}

void forint::
save_offsets(offsets_t& r)
{
  r.ra = r_action.size();
  r.sa = s_action.size();
  r.num = number.size();
  r.var = var_num.size();
  r.poin = poin_num.size();
  r.poin_level = poin_num_level.size();
  r.func0 = usr_function_0_num.size();
  r.func1 = usr_function_1_num.size();
  r.func2d = usr_function_2d_num.size();
  r.func2p = usr_function_2p_num.size();
  r.func2 = usr_function_2_num.size();
  r.offsets = offsets.size();
}

void forint::
subtr_offsets(offsets_t& r)
{
  r.ra      = r_action.size() - r.ra;
  r.sa      = s_action.size() - r.sa;
  r.num     = number.size() - r.num;
  r.var     = var_num.size() - r.var;
  r.poin    = poin_num.size() - r.poin;
  r.poin_level = poin_num_level.size() - r.poin_level;
  r.func0   = usr_function_0_num.size() - r.func0;
  r.func1   = usr_function_1_num.size() - r.func1;
  r.func2d  = usr_function_2d_num.size() - r.func2d;
  r.func2p  = usr_function_2p_num.size() - r.func2p;
  r.func2   = usr_function_2_num.size() - r.func2;
  r.offsets = offsets.size() - r.offsets + 1;
}

void forint::
inverse_subtr_offsets(offsets_t& r)
{
  r.ra      -= r_action.size();
  r.sa      -= s_action.size();
  r.num     -= number.size();
  r.var     -= var_num.size();
  r.poin    -= poin_num.size();
  r.poin_level -= poin_num_level.size();
  r.func0   -= usr_function_0_num.size();
  r.func1   -= usr_function_1_num.size();
  r.func2d  -= usr_function_2d_num.size();
  r.func2p  -= usr_function_2p_num.size();
  r.func2   -= usr_function_2_num.size();
  r.offsets -= offsets.size();
}

// skip sequential "ch" characters,
// WARNING: end of skip either non-space or '\0' symbol
void forint::
skip(char*& str)
{ while (*str && *str == ' ') ++str; }

void forint:: // dangerous function
rev_skip(char*& str)
{ while(*str == ' ') --str; }

// return: > 0 and force=false if error,
//         the number of element + 1 if there is one,
//         0 if there is no such element
int forint::
check_name(const char* name,
	   const char* mess,
	   bool        func_flag,
	   bool&       force)
{
  using namespace FInt_internal;
  const char* nm = name;
  char** p  = 0;
  if (isdigit(*name)) {
    c.err(0,0,mess);
if (!locale) {
    c.err("Name \'");
    c.err(name,5);
    c.err("\' is wrong.\n");
    c.warn("It must start with an alphabetic character"
	   " or the underscore.\n");
} else {
    c.err(" \'");
    c.err(name,5);
    c.err("\' .\n");
    c.warn("      ޣ.\n");
}
    c.flsh(); 
    force = false;
    error_code = 2;
    return error_code;
  }
  while (*nm) {
    if (!isalnum(*nm) &&
	*nm != '_') {
      c.err(0,0,mess);
if (!locale) {
      c.err("Name \'");
      c.err(name,5);
      c.err("\' is wrong.\n");
      c.warn("It can consist of alphabetic and numeric "
	     "characters and underscore.\n");
} else {
      c.err(" \'");
      c.err(name,5);
      c.err("\' .\n");
      c.warn("    ,   ޣ.\n");
}
      c.flsh(); 
      force = false;
      error_code = 3;
      return error_code;
    }
    ++nm;
  }
  if (func_flag) {
    for (uint i=(uint)FI_SIN; i<(uint)FI_NOT; i++)
      if (!strcmp(name,cmd[i])) {
	c.warn(0,0,mess);
if (!locale) {
	c.warn("Built-in function \'");
	c.warn(name,5);
	c.warn("()\' can't be overloaded for a DLF mode.\n");
} else {
	c.warn("  \'");
	c.warn(name,5);
	c.warn("()\'      DLF .\n");
}
        c.flsh(); 
      }
    for (uint i=(uint)FI_ATAN2; i<(uint)FI_COND; i++)
      if (!strcmp(name,cmd[i])) {
	c.warn(0,0,mess);
if (!locale) {
	c.warn("Built-in function \'");
	c.warn(name,5);
	c.warn("()\' can't be overloaded for a DLF mode.\n");
} else {
	c.warn("  \'");
	c.warn(name,5);
	c.warn("()\'      DLF .\n");
}
        c.flsh(); 
      }
    p = usr_function_0_name.getPtr();
    for (uint i=0; i<usr_function_0_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.warn("  \'");
	  c.warn(name,5);
	  c.warn("()\'      DLF .\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn("  \'");
	  c.warn(name,5);
	  c.warn("()\'      DLF .\n");
}
          c.flsh(); 
	  error_code = 3000;
	  return i+1;
	}
      }
    p = usr_function_1_name.getPtr();
    for (uint i=0; i<usr_function_1_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.err(" \'");
	  c.err(name,5);
	  c.err("\'     forint.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn(" \'");
	  c.warn(name,5);
	  c.warn("\'  .\n");
}
          c.flsh(); 
	  error_code = 3000;
	  return i+1;
	}
      }
    p = usr_function_2d_name.getPtr();
    for (uint i=0; i<usr_function_2d_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.err(" \'");
	  c.err(name,5);
	  c.err("\'     forint.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn(" \'");
	  c.warn(name,5);
	  c.warn("\'  .\n");
}
          c.flsh(); 
	  error_code = 3000;
	  return i+1;
	}
      }
    p = usr_function_2p_name.getPtr();
    for (uint i=0; i<usr_function_2p_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.err(" \'");
	  c.err(name,5);
	  c.err("\'     forint.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn(" \'");
	  c.warn(name,5);
	  c.warn("\'  .\n");
}
          c.flsh(); 
	  error_code = 3000;
	  return i+1;
	}
      }
    p = usr_function_2_name.getPtr();
    for (uint i=0; i<usr_function_2_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.err(" \'");
	  c.err(name,5);
	  c.err("\'     forint.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn(" \'");
	  c.warn(name,5);
	  c.warn("\'  .\n");
}
          c.flsh(); 
	  error_code = 3000;
	  return i+1;
	}
      }
  } else {
    p = var_name.getPtr();
    for (uint i=0; i<var_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.err(" \'");
	  c.err(name,5);
	  c.err("\'     forint.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn(" \'");
	  c.warn(name,5);
	  c.warn("\'  .\n");
}
	  error_code = 3000;
	  return i+1;
	}
      }
    p = poin_name.getPtr();
    for (uint i=0; i<poin_name.size(); i++)
      if (!strcmp(name,p[i])) {
	if (!force) {
	  c.err(0,0,mess);
if (!locale) {
	  c.err("There is name \'");
	  c.err(name,5);
	  c.err("\' in a forint dictionary already.\n");
} else {
	  c.err(" \'");
	  c.err(name,5);
	  c.err("\'     forint.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	else {
	  c.warn(0,0,mess);
if (!locale) {
	  c.warn("Name \'");
	  c.warn(name,5);
	  c.warn("\' will be reloaded.\n");
} else {
	  c.warn(" \'");
	  c.warn(name,5);
	  c.warn("\'  .\n");
}
          c.flsh(); 
	  error_code = 3000;
	  return i+1;
	}
      }
  }
  return 0;
}

int forint::
check_type(const char*& type,
	   b_type_t& b_type,
	   uint& level,
	   const char* mess)
{
  using namespace FInt_internal;
  char* tp = (char*)type;
  while (isalpha(*tp)) ++tp;
  uint size = tp - type;

  char* tb = tp;
  while (*tb == ' ') ++tb;
  char* te = tb;
  while (isalpha(*te)) ++te;
  uint sizz = te - tb;

  level = 0;
  if (!strncmp(type,"char",size))
    b_type = FI_CHAR;
  else if (!strncmp(type,"short",size))
    b_type = FI_SHORT;
  else if (!strncmp(type,"int",size))
    b_type = FI_INT;
  else if (!strncmp(type,"long",size)) {
    b_type = FI_LONG;
    if (!strncmp(tb,"long",sizz)) {
      b_type = FI_LONG_LONG;
      tp = te;
    }
    if (!strncmp(tb,"double",sizz)) {
      b_type = FI_LONG_DOUBLE;
      tp = te;
    }
  }
  else if (!strncmp(type,"float",size))
    b_type = FI_FLOAT;
  else if (!strncmp(type,"double",size))
    b_type = FI_DOUBLE;
  else {
    /* old branch
    c.err(0,0,mess);
    c.err("Unknown basic type \'");
    c.err(type);
    c.err("\'.");
    c.warn(" It must be from the set: ");
    c.warn("char"); c.warn(", ");
    c.warn("short"); c.warn(", ");
    c.warn("int"); c.warn(", ");
    c.warn("long"); c.warn(", ");
    c.warn("float"); c.warn(", ");
    c.warn("double"); c.warn(".\n");
    c.flsh(); 
    error_code = 6;
    return error_code;
    */
    b_type = FI_USR_CLASS;
  }
  while (*tp) {
    skip(tp);
    if (*tp == '*') {
      ++level;
      ++tp;
    }
    else {
      c.err(tp,1,mess);
if (!locale) {
      c.err("Wrong symbol in pointer type.\n");
} else {
      c.err("    .\n");
}
      c.flsh(); 
      error_code = 7;
      return error_code;
    }
  }
  return 0;
}

int forint::
putConst(const double& cnst,
	 const char* name,
	 bool force=false)
{
  error_code = 0;
  int rv = 0;
  rv=putVar((double&)cnst, name, force);
  if (!rv) {
    var_const_flag.pop_back();
    var_const_flag.push_back(true);
  }
  else if (force) {
    var_const_flag.replace(rv-1,true);
    return 0;
  }
  return rv;
}

int forint::
putVar(double& var,
       const char* name,
       bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  error_code = 0;
  c.setStr(name,strlen(name));

  if ((rv=check_name(name,"putVar(): ",false,force)) && !force)
    return rv;
  if (!rv) {
    var_name.push_back(strdup(name));
    variable.push_back(&var);
    var_const_flag.push_back(false);
  }
  else {
    free(var_name.getPtr()[rv-1]);
    var_name.replace(rv-1,strdup(name));
    variable.replace(rv-1,&var);
    var_const_flag.replace(rv-1,false);
  }
  variable_p = variable.getPtr();
  return 0;
}

int forint::
putPoin(void* ptr,
	const char* name,
	const char* type,
	vector<int>* dim_size,
	bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  b_type_t bt;
  uint  lv;
  error_code = 0;

  c.setStr(type,strlen(type));
  if ((rv=check_name(name,"putPoin(): ",false,force)) && !force)
    return rv;
  string nn = (string)"putPoin(), name = \'" + (string)name + (string)"\': ";
  if ((rv=check_type(type,bt,lv,(char*)nn.c_str())))
    return rv;
  if (bt == FI_USR_CLASS) {
    if (!lv) {
      c.err(0,0,"\'");
      c.err(type,5);
if (!locale) {
      c.err("\' isn't a pointer to a user class.\n");
} else {
      c.err("\'    -  .\n");
}
      c.flsh(); 
      error_code = 50;
      return error_code;
    }
  }
  /*
  if (dim_size && !dim_size->size()) {
    c.err(0,0,"putPoin(), name = \'");
    c.err(name);
    c.err("\': ");
    c.err("the number of dimensions is equal to zero.");
    c.flsh(); 
    error_code = 12;
    return error_code;
  }
  */
// realy only dim_size->size()-1 numbers are interested for us but to simplify
// an initiation dim_size for a programmer 
  if (!rv) {
    poin_name.push_back(strdup(name));
    pointer.push_back(ptr);
  }
  else {
    free(poin_name.getPtr()[rv-1]);
    poin_name.replace(rv-1,strdup(name));
    pointer.replace(rv-1,ptr);
  }
  Llong offset = 1;
  uint* tmp = (dim_size)
    ? (uint*)malloc(dim_size->size()*sizeof(uint))
    : 0;
  uint* tmp2 = (dim_size)
    ? (uint*)malloc(dim_size->size()*sizeof(uint))
    : 0;
// because dimension sizes are kept in such a way:
//   let W[a][b][c] is an array, A, B, C -- dimension sizes for a, b and c
//   indices. Then dim_size keeps numbers in the order: A, B, C.
  if (dim_size)
    for (int i=dim_size->size()-1; i>=0; i--) {
      tmp[i] = offset;
      offset *= (*dim_size)[i];
      tmp2[i] = (*dim_size)[i] - 1;
    }
  if (!rv) {
    poin_offset.push_back(tmp);
    poin_dim_size.push_back(tmp2);
  }
  else {
    free(poin_offset.getPtr()[rv-1]);
    free(poin_dim_size.getPtr()[rv-1]);
    poin_offset.replace(rv-1,tmp);
    poin_dim_size.replace(rv-1,tmp2);
  }
  poin_t pt;
  pt.poin_type = type;
  pt.b_type = bt;
  pt.level  = lv;
  pt.num_dims = (dim_size)? dim_size->size() : 0;
  if (dim_size && dim_size->size()) {
    pt.dim_size = (uint*)malloc(dim_size->size()*sizeof(uint));
    for (uint i=0; i<pt.num_dims; i++)
      pt.dim_size[i] = (*dim_size)[i];
  }
  if (!rv)
    poin_type.push_back(pt);
  else {
    free(poin_type.getPtr()[rv-1].dim_size);
    poin_type.replace(rv-1,pt);
  }
  pointer_p = pointer.getPtr();
  return 0;
}

int forint::
putFunc0(USR_FUNC_0_PTR ptr,
	 const char* name,
	 bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  error_code = 0;
  c.setStr(name,strlen(name));

  if ((rv=check_name(name,"putFunc0(): ",true,force)) && !force)
    return rv;
  if (!rv) {
    usr_function_0_name.push_back(strdup(name));
    usr_function_0.push_back(ptr);
  }
  else {
    free(usr_function_0_name.getPtr()[rv-1]);
    usr_function_0_name.replace(rv-1,strdup(name));
    usr_function_0.replace(rv-1,ptr);
  }
  usr_function_0_p = usr_function_0.getPtr();
  return 0;
}

int forint::
putFunc1(USR_FUNC_1_PTR ptr,
	 const char* name,
	 bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  error_code = 0;
  c.setStr(name,strlen(name));

  if ((rv=check_name(name,"putFunc1(): ",true,force)) && !force)
    return rv;
  if (!rv) {
    usr_function_1_name.push_back(strdup(name));
    usr_function_1.push_back(ptr);
  }
  else {
    free(usr_function_1_name.getPtr()[rv-1]);
    usr_function_1_name.replace(rv-1,strdup(name));
    usr_function_1.replace(rv-1,ptr);
  }
  usr_function_1_p = usr_function_1.getPtr();
  return 0;
}

int forint::
putFunc2D(USR_FUNC_2D_PTR ptr,
	  const char* name,
	  bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  error_code = 0;
  c.setStr(name,strlen(name));

  if ((rv=check_name(name,"putFunc2D(): ",true,force)) && !force)
    return rv;
  if (!rv) {
    usr_function_2d_name.push_back(strdup(name));
    usr_function_2d.push_back(ptr);
  }
  else {
    free(usr_function_2d_name.getPtr()[rv-1]);
    usr_function_2d_name.replace(rv-1,strdup(name));
    usr_function_2d.replace(rv-1,ptr);
  }
  usr_function_2d_p = usr_function_2d.getPtr();
  return 0;
}

int forint::
putFunc2P(USR_FUNC_2P_PTR ptr,
	  const char* name,
	  const char* type1,
	  const char* type2,
	  bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  b_type_t bt;
  uint  lv;
  error_code = 0;
  c.setStr(name,strlen(name));

  if ((rv=check_name(name,"putFunc2P(): ",true,force)))
    return rv;
  if (!rv) {
    usr_function_2p_name.push_back(strdup(name));
    usr_function_2p.push_back(ptr);
  }
  else {
    free(usr_function_2p_name.getPtr()[rv-1]);
    usr_function_2p_name.replace(rv-1,strdup(name));
    usr_function_2p.replace(rv-1,ptr);
  }
  string nn = (string)"putFunc2P(), name = \'" + (string)name +
    (string)"\': ";
  c.setStr(type1,strlen(type1));
  if ((rv=check_type(type1,bt,lv,(char*)nn.c_str())))
    return rv;
  if (bt == FI_USR_CLASS) {
    if (!lv) {
      c.err(0,0,"\'");
      c.err(type1,5);
if (!locale) {
      c.err("\' isn't a pointer to a user class (first argument of \'\n");
      c.err(name,5);
      c.err("\' function).\n");
} else {
      c.err("\'    -   "
	    "(   \'\n");
      c.err(name,5);
      c.err("\').\n");
}
      c.flsh(); 
      error_code = 51;
      return error_code;
    }
  }
  poin_t pt1;
  pt1.poin_type = type1;
  pt1.b_type = bt;
  pt1.level  = lv;
  if (!rv)
    usr_function_2p_poin_type.push_back(pt1);
  else
    usr_function_2p_poin_type.replace(2*(rv-1),pt1);
  c.setStr(type2,strlen(type2));
  if ((rv=check_type(type2,bt,lv,(char*)nn.c_str())))
    return rv;
  if (bt == FI_USR_CLASS) {
    if (!lv) {
      c.err(0,0,"\'");
      c.err(type2,5);
if (!locale) {
      c.err("\' isn't a pointer to user class (second argument of \'\n");
      c.err(name,5);
      c.err("\' function).\n");
} else {
      c.err("\'    -   "
	    "(   \'\n");
      c.err(name,5);
      c.err("\').\n");
}
      c.flsh(); 
      error_code = 52;
      return error_code;
    }
  }
  poin_t pt2;
  pt2.poin_type = type2;
  pt2.b_type = bt;
  pt2.level  = lv;
  if (!rv)
    usr_function_2p_poin_type.push_back(pt2);
  else
    usr_function_2p_poin_type.replace(2*(rv-1)+1,pt2);
  usr_function_2p_p = usr_function_2p.getPtr();
  return 0;
}

int forint::
putFunc2(USR_FUNC_2_PTR ptr,
	 const char* name,
	 const char* type,
	 bool force=false)
{
  error_code = 0;
  using namespace FInt_internal;
  int rv = 0;
  b_type_t bt;
  uint  lv;
  error_code = 0;
  c.setStr(name,strlen(name));

  c.setStr(type,strlen(type));
  if ((rv=check_name(name,"putFunc2(): ",true,force)) && !force)
    return rv;
  if (!rv) {
    usr_function_2_name.push_back(strdup(name));
    usr_function_2.push_back(ptr);
  }
  else {
    free(usr_function_2_name.getPtr()[rv-1]);
    usr_function_2_name.replace(rv-1,strdup(name));
    usr_function_2.replace(rv-1,ptr);
  }
  string nn = (string)"putFunc2(), name = \'" + (string)name +
    (string)"\': ";
  if ((rv=check_type(type,bt,lv,(char*)nn.c_str())))
    return rv;
  if (bt == FI_USR_CLASS) {
    if (!lv) {
      c.err(0,0,"\'\n");
      c.err(type,5);
if (!locale) {
      c.err("\' isn't a pointer to user class (second argument of \'\n");
      c.err(name,5);
      c.err("\' function).\n");
} else {
      c.err("\'    -   "
	    "(   \'\n");
      c.err(name,5);
      c.err("\').\n");
}
      c.flsh(); 
      error_code = 53;
      return error_code;
    }
  }
  poin_t pt;
  pt.poin_type = type;
  pt.b_type = bt;
  pt.level  = lv;
  if (!rv)
    usr_function_2_poin_type.push_back(pt);
  else
    usr_function_2_poin_type.replace(rv-1,pt);
  usr_function_2_p = usr_function_2.getPtr();
  return 0;
}

// return zero if the pair "rb" for an "lb" was found...
int forint::
pair(char*& str,
     uint   size,
     char   lb,
     char   rb)
{
  char* strb = str;
  uint  ss = size;
  uint counter = 1;
  /*  i've commented out it for user defined functions with empty signature
  if (*str == rb) {
    c.err(str,1,"Parse error.");
    error_code = 1;
    return error_code;
  }
  else
  */
  if (*str == rb)
    --counter;
  else if (*str == lb)
    ++counter;
  if (*str == ';' && counter) {
if (!locale) {
    c.err(str,1,"Parse error, semicolon in brackets.\n");
} else {
    c.err(str,1," ,     .\n");
}
    c.flsh(); 
    error_code = 98;
    return error_code;
  }
  while (counter && *str && size--) {
    ++str;
    if (*str == lb)
      ++counter;
    else if (*str == rb)
      --counter;
    if (*str == ';' && counter) {
if (!locale) {
      c.err(str,1,"Parse error, semicolon in brackets.\n");
} else {
      c.err(str,1," ,     .\n");
}
      c.flsh(); 
      error_code = 99;
      return error_code;
    }
  }
  if (!counter)
    return 0;
  else {
if (!locale) {
    c.err(strb-1,ss+1,"Unclosed left \'");
    c.err(lb,1,5);
    c.err("\' bracket.\n");
} else {
    c.err(strb-1,ss+1,"  \'");
    c.err(lb,1,5);
    c.err("\'  .\n");
}
    c.flsh(); 
    error_code = 1;
    return error_code;
  }
  return error_code;
}

// return zero if the pair "rb" for an "lb" was found...
int forint::
rev_pair(char*  beg,
	 char*& str,
	 char   lb,
	 char   rb)
{
  uint size = str-beg;
  uint ss = size;
  uint counter = 1;
  /*  i've commented out it for user defined functions with empty signature
  if (*str == rb) {
    c.err(str,1,"Parse error.");
    c.flsh(); 
    error_code = 1;
    return error_code;
  }
  else
  */
  if (str[0] == lb)
    --counter;
  else if (str[0] == rb)
    ++counter;
  if (*str == ';' && counter) {
if (!locale) {
    c.err(str,1,"Parse error, semicolon in brackets.\n");
} else {
    c.err(str,1," ,     .\n");
}
    c.flsh(); 
    error_code = 98;
    return error_code;
  }
  while (counter && size--) {
    --str;
    if (str[0] == lb)
      --counter;
    else if (str[0] == rb)
      ++counter;
    if (str[0] == ';' && counter) {
if (!locale) {
      c.err(str,1,"Parse error, semicolon in brackets.");
} else {
      c.err(str,1," ,     .\n");
}
      c.flsh(); 
      error_code = 98;
      return error_code;
    }
  }
  if (!counter)
    return 0;
  else {
    //c.err(strb-1,ss+1,"Unclosed right \'");
if (!locale) {
    c.err(beg,ss+2,"Unclosed right \'");
    c.err(rb,1,5);
    c.err("\' bracket.\n");
} else {
    c.err(beg,ss+2,"  \'");
    c.err(rb,1,5);
    c.err("\'  .\n");
}
    c.flsh(); 
    error_code = 1;
    return error_code;
  }
  return error_code;
}

// return true if there is ':' in a string str
// or false in the other case...
bool forint::
iftoken (char*  str,
	 char*  end,
	 char*& pos)
{
  uint coun = 1;
  while (*str && str!=end) {
    skip(str);
    if (*str == '(') {
      uint s = end-str-1;
      if (pair(++str,s,'(',')'))
	return false;
    }
    else if (*str == '[') {
      uint s = end-str-1;
      if (pair(++str,s,'[',']'))
	return false;
    }
    else if (*str == '?') {
      ++coun;
    }
    else if (*str == ':' && !(--coun)) {
      pos = str;
      return true;
    }
    ++str;
  }
  return false;
}

// return true if string "str" contains a number
// or false in the other case...
bool forint::
ntoken (char*  str,
	uint   size,
	char*& pos)
{
  char* str_b = str;
  using namespace FInt_internal;
  base = 10;
  skip(str);
  pos = str;
  size -= (str-str_b);
  bool first = true;
  bool exp_flag = false;

  while (size-- && *str) {
    if (first && size && str[0] == '0' && str[1] != '.') {
      base = 8;
      if (size > 1 && str[1] == 'x')
	base = 16;
      if (size > 2 &&
	  isdigit(str[1]) && str[2] == 'x') {
	string tmp (str,2);
	char* e;
	if (!(base=strtol(tmp.c_str(),&e,10)))
	  return false;
	else {
	  pos += 3;
	  size -= 2;
	  str += 2;
	}
      }
      /* old random base notation:
      // Literals  in other bases have prefix `0Nx', where N - number's base.
      else if (size > 3 &&
	       isdigit(str[1]) && isdigit(str[2]) && str[3] == 'x') {
	string tmp (str,3);
	char* e;
	if (!(base=strtol(tmp.c_str(),&e,10)))
	  return false;
	else {
	  pos += 4;
	  size -= 3;
	  str += 3;
	  }
      }
      */
      first = false;
      ++str;
      continue;
    }
    if (isdigit(str[0])) {
      ++str;
      if (first) first = false;
      continue;
    }
    if ((*str == '+' || *str == '-') && first) {
      first = false;
      skip(str);
      ++str;
      continue;
    }
    if (str[0] == '.' && !exp_flag &&
	(size > 0 && isdigit(str[1]) || !first && isdigit(str[-1]))) {
      first = false;
      ++str;
      continue;
    }
    if(!first && size > 0 &&
       (str[0]=='e' || str[0]=='E') &&
       isdigit(str[-1]) &&
       isdigit(str[1])) {
      str += 1;
      size -= 1;
      first = false;
      exp_flag = true;
      ++str;
      continue;
    }
    if(!first && size > 1 &&
       (str[0]=='e' || str[0]=='E') &&
       isdigit(str[-1]) &&
       (str[1]=='+' || str[1]=='-') && isdigit(str[2])) {
      str += 2;
      size -= 2;
      first = false;
      exp_flag = true;
      ++str;
      continue;
    }
    if (!first && *str == ' ') {
      ++str;
      continue;
    }
    if (base > 10 && isalpha(*str)) {
      ++str;
      continue;
    }
    if (!first && size && *str == '#') {
      string tmp (str_b,str-str_b);
      char* e;
      if (!(base=strtol(tmp.c_str(),&e,10)))
	return false;
      pos = ++str;
      continue;
    }
    return false;
  }
  return true;
}

// is "t" an operator delimeter...
bool forint::
is_op_delim(char t)
{
  if (t == ' ' ||
      t == '_' ||
      isalnum(int(t)))
    return true;
  return false;
}

// only for operators !!!
// return true if there is "t" in a string str
// or false in the other case...
bool forint::
optoken (char*  str,
	 uint   size,
	 char*  t,
	 char*& pos)
{
  if (!(*str)) return false;
  bool  rv       = false;
  char* bv       = str;
  bool  non_sign = false;
  bool  first    = true;
  while (size--) {
    if (*str == ' ') {
      ++str;
      continue;
    }
    if (first &&
	*str != '+' && *str != '-') {
      non_sign = true;
      first = false;
    }
// skip a delimeter...
    if (is_op_delim(*str)) {
      ++str;
      continue;
    }
    if (*str=='(') {
      char* lp = str;
      if (pair(++str,size,'(',')')) {
	pos = 0;
	return false;
      }
      size -= (str-lp);
      ++str;
      continue;
    }
    if (*str=='[') {
      char* lp = str;
      if (pair(++str,size,'[',']')) {
	pos = 0;
	return false;
      }
      size -= (str-lp);
      ++str;
      continue;
    }
    pos = str;
    if (!strncmp(str,t,strlen(t))) {
      rv = true;
      if ((t == "+" || t == "-") && !non_sign) // is it a sign
	rv = false;
// is it order's sign or number's sign ?...
      if ((t == "+" || t == "-") &&
	  str != bv &&
	  (str[-1] == 'e' || str[-1] == 'E') &&
	  isdigit(str[1])) // a sign
	rv = false;
// is the next symbol an operator delimeter ?...
      char* n_s = pos+strlen(t);
      uint  n_s_c = size - strlen(t) + 1;
      while (n_s_c-- && *n_s == ' ') n_s++;
      if (!is_op_delim(*n_s) &&
	  *n_s != '(' && *n_s != '+' && *n_s != '-' && *n_s != '.' &&
	  *n_s != '!' && *n_s != '~')
	rv = false;
// is the prev symbol an operator delimeter ?...
      char* p_s = pos-1;
      uint  p_s_c = (pos-bv);
      while (p_s_c-- && *p_s == ' ') p_s--;
      if (!is_op_delim(*p_s) && *p_s != ')' && *p_s != ']')
	rv = false;
      if (rv) return rv;
    }
    ++str;
  }
  pos = (rv) ? pos : 0;
  return rv;
}

// only for the highest priority !!!
// return number of an user pointer if string "str" contains a token bound
// with an element tok or 0 in the other case...
action_t forint::
token (char*    str,
       uint     size,
       action_t tok,
       char*    ldelim,
       char*    rdelim,
       char*&   pos,
       uint&    numb)
{
  using namespace FInt_internal;
  char* tok_beg = str;
  uint  token_size = 0;
  if (!size) return FI_COOR;
  uint  rest = size;
  if (!*str)
    return FI_COOR;
  for (uint j=0; j<strlen(ldelim); j++) {
    if (*str != ldelim[j]) {
      tok_beg = str;
      ++str;
      --rest;
      break;
    }
    ++str;
    --rest;
  }
  if (!rest)
    token_size = str - tok_beg;
  else
    while (rest) {
      for (uint j=0; j<strlen(rdelim); j++) {
	if (*str == rdelim[j] || !*str) {
	  token_size = str - tok_beg;
	  break;
	}
	if (rest == 1) {
	  token_size = str - tok_beg + 1;
	  str++;
	  break;
	}
      }
      if (token_size) break;
      ++str;
      --rest;
    }
// next (after a token) symbol-delimeter
  pos = str;
  if (token_size) {
    switch (tok) {
    case FI_VRBL:
      for (uint i=0; i<var_name.size(); i++) {
	if (!strncmp(tok_beg,var_name.getPtr()[i],token_size) &&
	    strlen(var_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_VRBL;
	}
      }
      break;

    case FI_POIN:
      for (uint i=0; i<poin_name.size(); i++) {
	if (!strncmp(tok_beg,poin_name.getPtr()[i],token_size) &&
	    strlen(poin_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_POIN;
	}
      }
      break;

    case FI_FUNC:
      for (uint i=0; i<usr_function_0_name.size(); i++) {
	if (!strncmp(tok_beg,usr_function_0_name.getPtr()[i],token_size) &&
	    strlen(usr_function_0_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_FUNC;
	}
      }
      break;

    case FI_FUNC1:
      for (uint i=0; i<usr_function_1_name.size(); i++) {
	if (!strncmp(tok_beg,usr_function_1_name.getPtr()[i],token_size) &&
	    strlen(usr_function_1_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_FUNC1;
	}
      }
      break;

    case FI_FUNC2D:
      for (uint i=0; i<usr_function_2d_name.size(); i++) {
	if (!strncmp(tok_beg,usr_function_2d_name.getPtr()[i],token_size) &&
	    strlen(usr_function_2d_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_FUNC2D;
	}
      }
      break;

    case FI_FUNC2P:
      for (uint i=0; i<usr_function_2p_name.size(); i++) {
	if (!strncmp(tok_beg,usr_function_2p_name.getPtr()[i],token_size) &&
	    strlen(usr_function_2p_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_FUNC2P;
	}
      }
      break;

    case FI_FUNC2:
      for (uint i=0; i<usr_function_2_name.size(); i++) {
	if (!strncmp(tok_beg,usr_function_2_name.getPtr()[i],token_size) &&
	    strlen(usr_function_2_name.getPtr()[i]) == token_size) {
	  numb = i;
	  return FI_FUNC2;
	}
      }
      break;

    default:;
    }
  }
  return FI_COOR;
}

// only for the highest priority !!!
// return enum!=COOR if string "str" contains a token bound with an element
// from [tb,te) or COOR in the other case...
action_t forint::
token (char*    str,
       uint     size,
       action_t tb,
       action_t te,
       char*    ldelim,
       char*    rdelim,
       char*&   pos)
{
  using FInt_internal::cmd;
  char* tok_beg = str;
  uint  token_size = 0;
  if (!size) return FI_COOR;
  uint  rest = size;
  if (!*str)
    return FI_COOR;
  for (uint j=0; j<strlen(ldelim); j++) {
    if (*str != ldelim[j]) {
      tok_beg = str;
      ++str;
      --rest;
      break;
    }
    ++str;
    --rest;
  }
  if (!rest)
    token_size = str - tok_beg;
  else
    while (rest) {
      for (uint j=0; j<strlen(rdelim); j++) {
	if (*str == rdelim[j] || !*str) {
	  token_size = str - tok_beg;
	  break;
	}
	if (rest == 1) {
	  token_size = str - tok_beg + 1;
	  str++;
	  break;
	}
      }
      if (token_size) break;
      ++str;
      --rest;
    }
// next (after a token) symbol-delimeter
  pos = str;
  if (token_size)
    for (uint i=tb; i<(uint)te; i++) {
      if (!strncmp(tok_beg,cmd[i],token_size) &&
	  strlen(cmd[i]) == token_size) {
	return (action_t)i;
      }
    }
  return FI_COOR;
}

char* forint::
get_non_zero_min(char* p1, char* p2)
{
  p1 = (p1==(char*)0x1) ? 0 : p1;
  if (!p1 && p2)
    return p2;
  if (p1 && !p2)
    return p1;
  if (p1 && p2)
    return (p1 < p2) ? p1 : p2;
  return (char*)1;
}

action_t forint::
find_token(uint   pr,     // priority
	   char*  str,    // string to check
	   uint   size,
	   char*& pos,    // position
	   uint&  numb)
{
  action_t rv = FI_COOR;
  char* p1,* p2,* p3,* p4;
  uint     ss = size;
  switch (pr) {
  case 2:
    if (optoken(str,size,"||",pos))
      return FI_OR;
    break;
  case 3:
    if (optoken(str,size,"&&",pos))
      return FI_AND;
    break;
  case 4:
    if (optoken(str,size,"|",pos))
      return FI_BOR;
    break;
  case 5:
    if (optoken(str,size,"^",pos))
      return FI_BXOR;
    break;
  case 6:
    if (optoken(str,size,"&",pos))
      return FI_BAND;
    break;
  case 7:
    optoken(str,size,"==",p1);
    optoken(str,size,"!=",p2);
    pos = get_non_zero_min(p1,p2);
    if (pos == p1) return FI_EQ;
    if (pos == p2) return FI_NE;
    break;
  case 8:
    optoken(str,size,">=",p1);
    optoken(str,size,">",p2);
    optoken(str,size,"<=",p3);
    optoken(str,size,"<",p4);
    pos = get_non_zero_min(p1,p2);
    pos = get_non_zero_min(pos,p3);
    pos = get_non_zero_min(pos,p4);
    if (pos == p1) return FI_GE;
    if (pos == p2) return FI_GT;
    if (pos == p3) return FI_LE;
    if (pos == p4) return FI_LT;
    break;
  case 9:
    optoken(str,size,"<<",p1);
    optoken(str,size,">>",p2);
    pos = get_non_zero_min(p1,p2);
    if (pos == p1) return FI_LOFF;
    if (pos == p2) return FI_ROFF;
    break;
  case 10:
    optoken(str,size,"+",p1);
    optoken(str,size,"-",p2);
    pos = get_non_zero_min(p1,p2);
    if (pos == p1) return FI_ADD;
    if (pos == p2) return FI_SUBT;
    break;
  case 11:
    optoken(str,size,"*",p1);
    optoken(str,size,"/",p2);
    optoken(str,size,"%",p3);
    pos = get_non_zero_min(p1,p2);
    pos = get_non_zero_min(pos,p3);
    if (pos == p1) return FI_MULT;
    if (pos == p2) return FI_DIV;
    if (pos == p3) return FI_MOD;
    break;
  case 12:
    pos = str;
    while (ss--) {
      if (*pos == '(')
	return FI_L_PARENTHESIS;
      if (*pos == '!')
	return FI_NOT;
      if (*pos == '~')
	return FI_COMP;
      if (*pos == '-')
	return FI_UN_MINUS;
      if (*pos == '+')
	return FI_UN_PLUS;
      if (*pos != ' ')
	break;
      pos++;
    }
    if (ntoken(str,size,pos))
      return FI_NUMB;
    if ((rv=token(str,size,FI_FUNC,  " (["," (",pos,numb)) ||
	(rv=token(str,size,FI_FUNC1, " (["," (",pos,numb)) ||
	(rv=token(str,size,FI_FUNC2D," (["," (",pos,numb)) ||
	(rv=token(str,size,FI_FUNC2P," (["," (",pos,numb)) ||
	(rv=token(str,size,FI_FUNC2, " (["," (",pos,numb)) ||
	(rv=token(str,size,FI_VRBL,  " (["," )",pos,numb)) ||
	(rv=token(str,size,FI_POIN,  " (["," [",pos,numb)) ||
	(rv=token(str,size,FI_FUNC,FI_OPER, " (["," (",pos)) ||
	(rv=token(str,size,FI_ATAN2,FI_COND," (["," (",pos)) ||
	flag_of_built_in_var_access &&
	(rv=token(str,size,FI_COOR,FI_RL, " (["," )]",pos)) ||
	flag_of_extra_built_in_var_access &&
	(rv=token(str,size,FI_RL,FI_VRBL, " (["," )]",pos)) ||
	(rv=token(str,size,FI_EN_E,FI_FUNC, " (["," )]",pos)))
      return rv;

  default:;
  }
  return FI_COOR;
}

action_t forint::
find_if_token(char*  str,          // string to check
	      char*  end,
	      char*& pos,          // "if" position
	      char*& el_pos,       // "else" position
	      char*& cond_pos,
	      char*& cond_b)    // <condition> position
{
  skip(str);
  cond_b = str;
  if (str == end) return FI_COOR;
  while (str < end) {
    /*
    if (*str == '(') {
      tmp = str;
      skip(++str);
      cond_b = str; // condition beginning
      pair(str,end-str,'(',')');
      if (error_code)
	return FI_COOR;
      cond_size = str++ - cond_b; // condition end
    }
    */
    skip(str);
    if (str >= end) return FI_COOR;
    if (*str == '?') {
      if (str == cond_b) {
if (!locale) {
	c.err(str,1,"Branching without a condition.\n");
} else {
	c.err(str,1,"   .\n");
}
        c.flsh(); 
	error_code = 1;
	return FI_COOR;
      } else {
	pos = str++;
	if (str == end) {
if (!locale) {
	    c.err(pos,el_pos-pos,"Parse error.\n");
} else {
	    c.err(pos,el_pos-pos," .\n");
}
            c.flsh(); 
	    error_code = 1;
	    return FI_COOR;
	}
	if (iftoken(str,end,el_pos)) {
	  char* sk = str;
	  skip(sk);
	  if (sk == el_pos) {
if (!locale) {
	    c.err(pos,el_pos-pos,"Empty branch.\n");
} else {
	    c.err(pos,el_pos-pos," .\n");
}
            c.flsh(); 
	    error_code = 1;
	    return FI_COOR;
	  }
	  char* skk = el_pos+1;
	  skip(skk);
	  if (skk == end) {
if (!locale) {
	    c.err(el_pos+1,end-el_pos-1,"Empty branch.\n");
} else {
	    c.err(el_pos+1,end-el_pos-1," .\n");
}
            c.flsh(); 
	    error_code = 1;
	    return FI_COOR;
	  }
	  return FI_IF;
	} else {
if (!locale) {
	  c.err(pos+1,end-pos-1,"There is no \'");
	  c.err(":",5);
	  c.err("\' twin to \'");
	  c.warn("?",5);
	  c.err("\'.\n");
} else {
	  c.err(pos+1,end-pos-1," \'");
	  c.err(":",5);
	  c.err("\'  \'");
	  c.warn("?",5);
	  c.err("\'.\n");
}
          c.flsh(); 
	  error_code = 1;
	  return FI_COOR;
	}
      }
    }
    ++str;
  }
  return FI_COOR;
}

bool forint::
is_lvalue(char* beg,
	  char* end)
{
  using FInt_internal::cmd;
  skip(beg);
  char* tmp = beg;
  while (tmp!=end) {
    if (*tmp == '_' || isalnum(*tmp)) {
      ++tmp;
      continue;
    }
    if (*tmp == ' ' || *tmp == '[') {
      end = tmp;
      break;
    }
    return false;
  }
  uint s = end-beg;
  for (uint i=FI_X; i<FI_SELF; i++)
    if (!strncmp(beg,cmd[i],s) &&
	s == strlen(cmd[i]))
      return true;
  for (uint i=0; i<var_name.size(); i++)
    if (!strncmp(beg,var_name.getPtr()[i],s) &&
	s == strlen(var_name.getPtr()[i]) &&
	!var_const_flag.getPtr()[i])
      return true;
// there is no dereferencing checking (it is performed later)
  for (uint i=0; i<poin_name.size(); i++) {
      char* t = beg;
      while (*t != '[') ++t;
    if (!strncmp(beg,poin_name.getPtr()[i],t-beg) &&
	t-beg == abs(strlen(poin_name.getPtr()[i]))) {
      return true;
    }
  }
  return false;
}

// return true if assignment statement was found
bool forint::
get_assignment(char*     expr,
	       char*&    pos,
	       uint&     length,
	       action_t& cmnd)
{
  using FInt_internal::cmd;
  if (pos-expr > 0 &&
      (isalnum(*(pos-1)) ||
       *(pos-1) == ' '   ||
       *(pos-1) == ')'   ||
       *(pos-1) == ']') ) {
    length = 1;
    cmnd = FI_ASS;
    return true;
  } else if (pos-expr > 1 &&
	     (isalnum(*(pos-2)) ||
	      *(pos-2) == ' ' ||
	      *(pos-2) == ')' ||
	      *(pos-2) == ']') ) {
    for (uint i=FI_ASS_PL; i<FI_ASS_LOFF; i++)
      if (!strncmp(cmd[i],pos-1,2)) {
	pos -= 1;
	length = 2;
	cmnd = (action_t)i;
	return true;
      }
  } else if (pos-expr > 2 &&
	     (isalnum(*(pos-3)) ||
	      *(pos-3) == ' ' ||
	      *(pos-3) == ')' ||
	      *(pos-3) == ']') ) {
    if (!strncmp("<<=",pos-2,2)) {
      pos -= 2;
      length = 3;
      cmnd = FI_ASS_LOFF;
      return true;
    }	  
    if (!strncmp(">>=", pos-2, 2)) {
      pos -= 2;
      length = 3;
      cmnd = FI_ASS_ROFF;
      return true;
    }	  
  }  
  return false;
}
  
bool forint::
find_assignment(char*     expr,
		uint      size,
		char*&    pos,
		uint&     length,
		char*&    prev_pos,
		uint&     prev_length,
		action_t& cmnd)
{
  using FInt_internal::cmd;
  if (size>2) prev_pos = expr+size-1;
  else return false;
  prev_length = 0;
  bool first = false;
  action_t tmp = FI_COOR;

  while (prev_pos != expr) {
    if (*prev_pos == '=' &&
	prev_pos+1 != expr+size && *(prev_pos+1) != '=') {
      if (first) {
	if (get_assignment(expr, prev_pos, prev_length, tmp))
	  return true;
	else prev_length = 0;
      } else {
	pos = prev_pos;
	if (get_assignment(expr, pos, length, cmnd))
	  first = true;
      }
    }
    if (*prev_pos == ')' && prev_pos > expr)
      if (rev_pair(expr, --prev_pos, '(', ')'))
	return false;
      else continue;
    if (*prev_pos == ']' && prev_pos > expr)
      if (rev_pair(expr, --prev_pos, '[', ']'))
	return false;
      else continue;
    --prev_pos;
  }
  return (first) ? true : false;
}

int forint::
decompose(char* expr,     // expression to decompose
	  char* end,      // expr end (can't be used as an expr's symbol !!!)
	  int  priority,  // start to decompose with "priority"
	  uint& ccoun)    // the number of `FI_CONTINUE' operators
{
  using namespace FInt_internal;
  char* sk_expr = expr;
  while (*sk_expr == ' ' && sk_expr != end) sk_expr++;
  uint s = (sk_expr-expr) ? sk_expr-expr : 1;
  if (sk_expr == end) {
if (!locale) {
    c.err(expr,s,"Empty expression.\n");
} else {
    c.err(expr,s," .\n");
}
    c.flsh(); 
    error_code = 6;
    return error_code;
  }
  action_t command = FI_COOR;
  char* pos       = 0;
  char* prev_pos  = 0;
  char* el_pos    = 0;
  char* cond_pos  = 0;
  char* cond_b    = 0;
  char* tail = 0;
  char* ctmp = 0;
  uint tmp3       = 0;
  uint numb       = 0;
  uint size = end-expr;
  error_code = 0;
  if (!size) {
    error_code = -1;
    return error_code;
  }

  uint prev_numb;
  uint tmp_size;
  bool first = true;
  action_t comm;
  char* tm1;
  uint  tm2;
  char* tm3;
  uint  l_ccoun = 0;

  switch (priority) {

// find 'while' statement
  case -1:
  if (flag_of_ctrl_statement_access) {
    tm1 = sk_expr;
    while(*tm1) {
      if (*tm1 == 'w' &&
	  (tm1 == sk_expr || !isalnum(*(tm1-1)) || *(tm1-1) != '_') &&
	  tm1[1] && tm1[1] == 'h' &&
	  tm1[2] && tm1[2] == 'i' &&
	  tm1[3] && tm1[3] == 'l' &&
	  tm1[4] && tm1[4] == 'e') {
	tm3 = tm1;
	tm1 += strlen("while");
	skip(tm1);
	if (*tm1 == '(') {
	  if (!first) {
if (!locale) {
	    c.err(tm3,end-tm3,"the \'");
	    c.err("while",5);
	    c.err("\' statement can't be preceded by anything "
		  "in a simple expression.\n");
} else {
	    c.err(tm3,end-tm3,"\'");
	    c.err("while",5);
	    c.err("\'       .\n");
}
            c.flsh(); 
	    error_code = 499;
	  }
	  else
	    command = FI_WHILE;
	  break;
	}
      }
      ++tm1;
      first = false;
    }
    if (command == FI_WHILE) {
      is_while = true;
      r_action.push_back(FI_WHILE);
      cond_pos = ++tm1;
      pair(tm1,end-tm1,'(',')');
      offsets_t r;
      save_offsets(r);
      if (decompose(cond_pos, tm1, 0, ccoun))
	return error_code;
      s_action.push_back(FI_IF);
      tmp3 = offsets.size();
      offsets_t r2;
      save_offsets(r2);
      ctmp = tm1+1;
      while (*ctmp == ' ' && ctmp != end) ctmp++;
      if (ctmp != end) {
	if (decompose(tm1+1, end, 0, ccoun))
	  return error_code;
	inverse_subtr_offsets(r);
	r.ra -= 1; // just after FI_WHILE
	r.offsets -= 1;
	offsets.push(r);
	s_action.push_back(FI_WHILE);
	subtr_offsets(r2);
	offsets.insert(tmp3,r2);
      }
      else {
if (!locale) {
	c.err(sk_expr,end-sk_expr,"Empty body of the \'");
	c.err("while",5);
	c.err("\' statement.\n");
} else {
	c.err(sk_expr,end-sk_expr,"  \'");
	c.err("while",5);
	c.err("\' .\n");
}
        c.flsh(); 
	error_code = 489;
      }
      break;
    }
    if (error_code)
      return error_code;
  }

  case 0:
    tmp_size = size;
    first = true;
    while (tmp_size &&
	   find_assignment(expr, tmp_size,
			   pos, numb, prev_pos, prev_numb, command)) {
      r_action.push_back(command);
      if (first) {
// decompose right value...
	if(decompose(pos+numb, end, 1, ccoun))
	  return error_code;
	first = false;
	if (command == FI_ASS_MOD  ||
	    command == FI_ASS_AND  ||
	    command == FI_ASS_OR   ||
	    command == FI_ASS_XOR  ||
	    command == FI_ASS_LOFF ||
	    command == FI_ASS_ROFF)
	  push_cast_pos(pos+numb,end);
      }
      char* b = prev_pos+prev_numb;
      char* e = pos;
      skip(b);
      rev_skip(e);
      if (!is_lvalue(b, e)) {
if (!locale) {
	c.err(b, e-b,"Non-lvalue in assignment.\n");
} else {
	c.err(b, e-b," left-value   .\n");
}
        c.flsh(); 
	error_code = 300;
	return error_code;
      } else {
// decompose left value...
	if(decompose(b, e, 12, ccoun))
	  return error_code;
	if (prev_numb)
	  r_action.push_back(FI_CONTINUE);
	tmp_size = prev_pos - expr + prev_numb + 1;
	if (command == FI_ASS_MOD  ||
	    command == FI_ASS_AND  ||
	    command == FI_ASS_OR   ||
	    command == FI_ASS_XOR  ||
	    command == FI_ASS_LOFF ||
	    command == FI_ASS_ROFF)
	  push_long_name_pos(b,string(b,e-b));
      }
      s_action.push_back(command);
    }
    if (error_code)
      return error_code;
    if (command != FI_COOR)
      break;

  case 1:
    if (find_if_token(expr, end, pos, el_pos,
		      cond_pos, cond_b)) {
      r_action.push_back(FI_IF);
      //num = r_action.size();
      if(decompose(cond_b, pos, 0, ccoun))
	return error_code;
      s_action.push_back(FI_IF);
      tmp3 = offsets.size();
      offsets_t r;
      save_offsets(r);
      if (decompose(pos+1, el_pos, 1, ccoun))
	return error_code;
      s_action.push_back(FI_ELSE);
      subtr_offsets(r);
      offsets.insert(tmp3,r);
      tmp3 = offsets.size();
      save_offsets(r);
      if (decompose(el_pos+1, end, 1, ccoun))
	return error_code;
      subtr_offsets(r);
      offsets.insert(tmp3,r);
      return error_code;
    }
    if (error_code)
      return error_code;

  case 2:
    if (find_token(2, expr, size, pos, numb)) {
      r_action.push_back(FI_OR);
      if (decompose(expr, pos, 3, l_ccoun=1))
	return error_code;
      if (find_token(2, pos+2, size - (pos+2-expr), tm1, tm2))
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
      s_action.push_back(FI_JUMP);
      tmp3 = offsets.size();
      offsets_t r;
      save_offsets(r);
      if (decompose(pos+2, end, 2, l_ccoun=0))
	return error_code;
      s_action.push_back(FI_OR);
      subtr_offsets(r);
      offsets.insert(tmp3,r);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 3:
    if (find_token(3, expr, size, pos, numb)) {
      r_action.push_back(FI_AND);
      if (decompose(expr, pos, 4, l_ccoun=1))
	return error_code;
      if (find_token(3, pos+2, size - (pos+2-expr), tm1, tm2))
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
      s_action.push_back(FI_JUMP);
      tmp3 = offsets.size();
      offsets_t r;
      save_offsets(r);
      if (decompose(pos+2, end, 3, l_ccoun=0))
	return error_code;
      s_action.push_back(FI_AND);
      subtr_offsets(r);
      offsets.insert(tmp3,r);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 4:
    if (find_token(4, expr, size, pos, numb)) {
      r_action.push_back(FI_BOR);
      if (decompose(expr, pos, 5, l_ccoun=1))
	return error_code;
      char* tmp_pos;
      if (find_token(4, pos+1, size - (pos+1-expr), tmp_pos, numb)) {
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	push_cast_pos(expr,pos);
	push_cast_pos(pos+1,tmp_pos);
      } else {
	push_cast_pos(expr,pos);
	push_cast_pos(pos+1,end);
      }
      if (decompose(pos+1, end, 4, l_ccoun=0))
	return error_code;
      s_action.push_back(FI_BOR);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 5:
    if (find_token(5, expr, size, pos, numb)) {
      r_action.push_back(FI_BXOR);
      if (decompose(expr, pos, 6, l_ccoun=1))
	return error_code;
      char* tmp_pos;
      if (find_token(5, pos+1, size - (pos+1-expr), tmp_pos, numb)) {
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	push_cast_pos(expr,pos);
	push_cast_pos(pos+1,tmp_pos);
      } else {
	push_cast_pos(expr,pos);
	push_cast_pos(pos+1,end);
      }
      if (decompose(pos+1, end, 5, l_ccoun=0))
	return error_code;
      s_action.push_back(FI_BXOR);
      return error_code;
      ++ccoun += l_ccoun;
    }
    if (error_code)
      return error_code;

  case 6:
    if (find_token(6, expr, size, pos, numb)) {
      r_action.push_back(FI_BAND);
      if (decompose(expr, pos, 7, l_ccoun=1))
	return error_code;
      char* tmp_pos;
      if (find_token(6, pos+1, size - (pos+1-expr), tmp_pos, numb)) {
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	push_cast_pos(expr,pos);
	push_cast_pos(pos+1,tmp_pos);
      } else {
	push_cast_pos(expr,pos);
	push_cast_pos(pos+1,end);
      }
      if (decompose(pos+1, end, 6, l_ccoun=0))
	return error_code;
      s_action.push_back(FI_BAND);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 7:
    if ((command=find_token(7, expr, size, pos, numb))) {
      r_action.push_back(command);
      if (decompose(expr, pos, 8, l_ccoun=0))
	return error_code;
      comm = command;
      while ((command=find_token(7, pos+2, size-(pos+2-expr), tm1, tm2))) {
	tmp3 = offsets.size();
	if (decompose(pos+2, tm1, 8, l_ccoun=1))
	  return error_code;
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	r_action.push_back(command);
	s_action.push_back(comm);
	comm = command;
	pos = tm1;
      }
      decompose(pos+2, end, 8, l_ccoun=0);
      s_action.push_back(comm);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 8:
    if ((command=find_token(8, expr, size, pos, numb))) {
      r_action.push_back(command);
      if (decompose(expr, pos, 9, l_ccoun=0))
	return error_code;
      comm = command;
      while ((command=find_token(8, pos+strlen(cmd[command]),
				 size-(pos+strlen(cmd[command])-expr),
				 tm1, tm2))) {
	if (decompose(pos+strlen(cmd[command]), tm1, 9, l_ccoun=1))
	  return error_code;
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	r_action.push_back(command);
	s_action.push_back(comm);
	comm = command;
	pos = tm1;
      }
      decompose(pos+strlen(cmd[comm]), end, 9, l_ccoun=0);
      s_action.push_back(comm);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 9:
    if ((command=find_token(9, expr, size, pos, numb))) {
      r_action.push_back(command);
      if (decompose(expr, pos, 10, l_ccoun=0))
	return error_code;
      comm = command;
      while ((command=find_token(9, pos+2, size-(pos+2-expr), tm1, tm2))) {
	if (decompose(pos+2, tm1, 10, l_ccoun=1))
	  return error_code;
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	r_action.push_back(command);
	s_action.push_back(comm);
	comm = command;
	pos = tm1;
      }
      decompose(pos+2, end, 10, l_ccoun=0);
      char* tmp_pos;
      if (find_token(9, pos+2, size - (pos+2-expr), tmp_pos, numb)) {
	push_cast_pos(expr,pos);
	push_cast_pos(pos+2,tmp_pos);
      } else {
	push_cast_pos(expr,pos);
	push_cast_pos(pos+2,end);
      }
      s_action.push_back(comm);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 10:
    if ((command=find_token(10, expr, size, pos, numb))) {
      r_action.push_back(command);
      if (decompose(expr, pos, 11, l_ccoun=0))
	return error_code;
      comm = command;
      while ((command=find_token(10, pos+1, size-(pos+1-expr), tm1, tm2))) {
	if (decompose(pos+1, tm1, 11, l_ccoun=1))
	  return error_code;
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	r_action.push_back(command);
	s_action.push_back(comm);
	comm = command;
	pos = tm1;
      }
      decompose(pos+1, end, 11, l_ccoun=0);
      s_action.push_back(comm);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 11:
    if ((command=find_token(11, expr, size, pos, numb))) {
      r_action.push_back(command);
      if (decompose(expr, pos, 12, l_ccoun=0))
	return error_code;
      comm = command;
      while ((command=find_token(11, pos+1, size-(pos+1-expr), tm1, tm2))) {
	if (decompose(pos+1, tm1, 12, l_ccoun=1))
	  return error_code;
	while (l_ccoun--) r_action.push_back(FI_CONTINUE);
	r_action.push_back(command);
	s_action.push_back(comm);
	comm = command;
	pos = tm1;
      }
      decompose(pos+1, end, 12, l_ccoun=0);
      if (comm == FI_MOD) {
	char* tmp_pos;
	if (find_token(11, pos+1, size - (pos+1-expr), tmp_pos, numb)) {
	  push_cast_pos(expr,pos);
	  push_cast_pos(pos+1,tmp_pos);
	} else {
	  push_cast_pos(expr,pos);
	  push_cast_pos(pos+1,end);
	}
      }
// put position of double() casting (both operands of `/' operator
// may be of integer type and result in DLF mode will be casted to integer
// also in the case)
      if (comm == FI_DIV)
	d_push_cast_pos(expr,pos);

      s_action.push_back(comm);
      ++ccoun += l_ccoun;
      return error_code;
    }
    if (error_code)
      return error_code;

  case 12:
    pos = 0;
    if (!(command=find_token(12, expr, size, pos, numb))) {
      char* b = expr;
      skip(expr);
if (!locale) {
      c.err(expr,size-(expr-b),"Wrong token or wrong expression.\n");
} else {
      c.err(expr,size-(expr-b),
	    "    .\n");
}
      c.flsh(); 
      error_code = 10;
      return error_code;
    } else {
      if (command==FI_L_PARENTHESIS) {
	char* pe = pos+1;
	skip(pe);
	if (*pe == ')') {
if (!locale) {
	  c.err(pe,1,"Parse error.\n");
} else {
	  c.err(pe,1," .\n");
}
          c.flsh(); 
	  error_code = 1;
	  return error_code;
	}
	pair(pe,size-(pe-expr),'(',')');
	if (error_code)
	  return error_code;
	if (decompose(pos+1, pe, 0, ccoun))
	  return error_code;
	tail = pe + 1;
      } else if (command==FI_NOT) {
	r_action.push_back(FI_NOT);
	if (decompose(pos+1, end, 12, ccoun))
	  return error_code;
      } else if (command==FI_COMP) {
	r_action.push_back(FI_COMP);
	if (decompose(pos+1, end, 12, ccoun))
	  return error_code;
	push_cast_pos(pos+1,end);
      } else if (command==FI_UN_MINUS) {
	r_action.push_back(FI_UN_MINUS);
	if (decompose(pos+1, end, 12, ccoun))
	  return error_code;
      } else if (command==FI_UN_PLUS) {
	if (decompose(pos+1, end, 12, ccoun))
	  return error_code;
      }
      else {
// functions, pointers, coordinates, numbers...
	r_action.push_back(command);
	if (command == FI_NUMB) {
	  if (base==10) {
	    double dres;
	    int tmp_errno = errno;
	    errno = 0;
	    number.push_back((dres=strtod(pos,&tail)));
	    if (dres == HUGE_VAL || dres == -HUGE_VAL) {
if (!locale) {
	      c.err(pos,tail-pos,"Overflow (too big number).\n");
} else {
	      c.err(pos,tail-pos," (  ).\n");
}
              c.flsh(); 
	      error_code = 203;
	      return error_code;
	    }
	    if (errno == ERANGE)
if (!locale) {
  c.warn(pos,tail-pos,"Underflow (too small number).\n");
  c.flsh(); 
} else {
  c.warn(pos,tail-pos,"  (  ).\n");
  c.flsh(); 
}
	    else errno = tmp_errno;
	  }
	  else if (base) {
	    if (base < 2 || base > 36) {
if (!locale) {
	      c.err(expr,pos-expr,"Wrong number's base.\n");
	      c.warn("It must be in the interval [");
	      c.warn("2",5);
	      c.warn(",");
	      c.warn("36",5);
	      c.warn("].\n");
} else {
	      c.err(expr,pos-expr,"  .\n");
	      c.warn("     [");
	      c.warn("2",5);
	      c.warn(",");
	      c.warn("36",5);
	      c.warn("].\n");
}
              c.flsh(); 
	      error_code = 210;
	      return error_code;
	    }
	    long lres;
	    number.push_back(double(lres=strtol(pos,&tail,base)));
	    if (lres == LONG_MAX) {
if (!locale) {
	      c.err(pos,tail-pos,"Overflow (too big number).\n");
} else {
	      c.err(pos,tail-pos," (  ).\n");
}
              c.flsh(); 
	      error_code = 201;
	      return error_code;
	    }
	    if (lres == LONG_MIN) {
if (!locale) {
	      c.warn(pos,tail-pos,"Underflow (too small number).\n");
} else {
	      c.warn(pos,tail-pos,"  (  ).\n");
}
              c.flsh(); 
	      error_code = 202;
	      return error_code;
	    }
	  }
// It will be hardly...
	  if (tail && tail > end) {
if (!locale) {
	    c.err(pos,tail-pos,
		  "Too many symbols have been red for reading of a number.\n");
} else {
	    c.err(pos,tail-pos,
		  "       .\n");
}
            c.flsh(); 
	    return 1;
	  }
	} else if (command == FI_VRBL) {
	  var_num.push_back(numb);
	} else if (command == FI_POIN) {
	  if (poin_type.getPtr()[numb].b_type == FI_USR_CLASS) {
if (!locale) {
	    c.err(expr,end-expr,
		  "User class pointer is used as not a function argument.\n");
} else {
	    c.err(expr,end-expr,
		  "        .\n");
}
              c.flsh(); 
	      error_code = 41;
	      return error_code;
	  }
	  poin_num.push_back(numb);
// find square brackets...
	  char* b_a = pos;
	  uint  nb  = 0; // the number of dereferencings
	  char* b_b_a = 0;
	  while (b_a != end) {
	    if (*b_a == '[') {
	      b_b_a = b_a;
	      ++nb;
	      ++b_a;
	      char* bb_a = b_a;
	      pair(b_a,end-b_a,'[',']');
	      if (decompose(bb_a, b_a, 0, ccoun))
		return error_code;
	    } else if (*b_a != ' ') {
if (!locale) {
	      c.err(b_a,1,"Wrong expression.\n");
} else {
	      c.err(b_a,1," .\n");
}
              c.flsh(); 
	      error_code = 13;
	      return error_code;
	    }
	    if (nb > poin_type.getPtr()[numb].num_dims) {
if (!locale) {
	      c.err(expr,end-expr,"Too many \'[]\'s.\n");
} else {
	      c.err(expr,end-expr,"  \'[]\'.\n");
}
              c.flsh(); 
	      error_code = 14;
	      return error_code;
	    }
	    ++b_a;
	  }
	  if (nb < poin_type.getPtr()[numb].num_dims) {
if (!locale) {
	    c.err(expr,end-expr,"Too small \'[]\'s.\n");
} else {
	    c.err(expr,end-expr,"  \'[]\'.\n");
}
              c.flsh(); 
	      error_code = 15;
	      return error_code;
	  }
	  poin_num_level.push_back(nb);
	} else if (command < FI_FUNC) {
// to check tail of the expr after coord...
	  tail = pos;
	} else if (command >= FI_FUNC) {
	  skip(pos);
	  if (*pos=='(') {
	    char* b_pos = ++pos;
	    skip(pos);
	    char* p = pos;
	    char* cp = 0;
	    pair(p,size,'(',')');
	    if (error_code)
	      return error_code;
	    if (pos == p) {
// user defined function with empty signature...
	      if (command == FI_FUNC)
		usr_function_0_num.push_back(numb);
	      else {
		uint s = (p-b_pos) ? p-b_pos : 1;
if (!locale) {
		c.err(b_pos,s,"Empty signature of a function.\n");
} else {
		c.err(b_pos,s,"  .\n");
}
                c.flsh(); 
		error_code = 3;
		return error_code;
	      }
	    } else {
	      tail = p+1;
	      for (char* r=pos; r<p; r++) {
// comma can be in 
		if (*r == '(') {
		  pair(++r,size,'(',')');
		  if (error_code)
		    return error_code;
		}
		if (*r == '[') {
		  pair(++r,size,'[',']');
		  if (error_code)
		    return error_code;
		}
		if (*r == ',') {
		  cp = r;
		  break;
		}
	      }
	      if (cp && (command == FI_FUNC1 ||
			 command > FI_FUNC2 && command < FI_OPER)) {
if (!locale) {
		c.err(cp,1,
		      "Comma in signature of a one-argument function.\n");
} else {
		c.err(cp,1,
		      "    .\n");
}
                c.flsh(); 
		error_code = 4;
		return error_code;
	      } else if (!cp && (command == FI_FUNC2D ||
				 command == FI_FUNC2P ||
				 command == FI_FUNC2 ||
				 command > FI_OPER && command < FI_COND)) {
// "> OPER" is right because operators (+,-,%) can't be here
if (!locale) {
		c.err(pos,p-pos,
		      "There is no delimiting comma in signature of a "
		      "two-argument function.\n");
} else {
		c.err(pos,p-pos,
		      "    - .\n");
}
                c.flsh(); 
		error_code = 4;
		return error_code;
	      } else if (command == FI_FUNC1) {
		if (decompose(pos, p, 0, ccoun))
		  return error_code;
		usr_function_1_num.push_back(numb);
	      } else if (command == FI_FUNC2P) {
// find first generic pointer...
		char* pp = 0;
		uint  nn = 0;
		if (token(pos,cp-pos,
			  FI_POIN,  " ("," )[",pp,nn) != FI_POIN) {
if (!locale) {
		  c.err(pos,cp-pos,"Must be a pointer.\n");
} else {
		  c.err(pos,cp-pos,"  .\n");
}
                  c.flsh(); 
		  error_code = 5;
		  return error_code;
		}
		r_action.push_back(FI_POIN);
		poin_num.push_back(nn);
// find square brackets...
		char* b_a = pp;
		uint  nb  = 0; // the number of dereferencings
		while (b_a != cp) {
		  if (*b_a == '[') {
		    ++nb;
		    ++b_a;
		    char* bb_a = b_a;
		    pair(bb_a,cp-bb_a,'[',']');
		    if (decompose(b_a, bb_a, 0, ccoun))
		      return error_code;
		    b_a = bb_a + 1;
		  } else if (*b_a != ' ') {
if (!locale) {
		    c.err(b_a,1,"Wrong expression.\n");
} else {
		    c.err(b_a,1," .\n");
}
                    c.flsh(); 
		    error_code = 13;
		    return error_code;
		  }
		}
		s_action.push_back(FI_POIN);
		poin_num_level.push_back(nb);
		if (usr_function_2p_poin_type.getPtr()[numb*2].level >
		    poin_type.getPtr()[nn].level +
		    poin_type.getPtr()[nn].num_dims - nb) {
if (!locale) {
		  c.err(pos,cp-pos, "Too many \'[]\'s.\n");
} else {
		  c.err(pos,cp-pos, "  \'[]\'.\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
		if (usr_function_2p_poin_type.getPtr()[numb*2].level <
		    poin_type.getPtr()[nn].level +
		    poin_type.getPtr()[nn].num_dims - nb) {
if (!locale) {
		  c.err(pos,cp-pos, "Too small \'[]\'s.\n");
} else {
		  c.err(pos,cp-pos, "  \'[]\'.\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
		if (poin_type.getPtr()[nn].num_dims < nb) {
if (!locale) {
		  c.err(pos,cp-pos, "Unknown dimension size(s).\n");
} else {
		  c.err(pos,cp-pos, "  .\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
// check first argument basic type...
		if (usr_function_2p_poin_type.getPtr()[numb*2].b_type !=
		    FI_USR_CLASS &&
		    usr_function_2p_poin_type.getPtr()[numb*2].b_type !=
		    poin_type.getPtr()[nn].b_type ||
		    usr_function_2p_poin_type.getPtr()[numb*2].b_type ==
		    FI_USR_CLASS &&
		    usr_function_2p_poin_type.getPtr()[numb*2].poin_type !=
		    poin_type.getPtr()[nn].poin_type) {
if (!locale) {
		  c.err(pos,cp-pos,
			"Formal and actual argument types mismatch (\'");
} else {
		  c.err(pos,cp-pos,
			"       (\'");
}
		  if (usr_function_2p_poin_type.getPtr()[numb*2].b_type !=
		      FI_USR_CLASS)
		    c.err(str_type[usr_function_2p_poin_type.
				  getPtr()[numb].b_type],5);
		  else
		    c.err(usr_function_2p_poin_type.getPtr()
			  [numb*2].poin_type,5);
if (!locale) {
		  c.err("\' and \'");
} else {
		  c.err("\'  \'");
}
		  if (poin_type.getPtr()[nn].b_type != FI_USR_CLASS)
		    c.err(str_type[poin_type.getPtr()[nn].b_type],5);
		  else
		    c.err(poin_type.getPtr()[nn].poin_type,5);
		  c.err("\').\n");
                  c.flsh(); 
		  error_code = 12;
		  return error_code;
		}
// find second generic pointer...
		pp = 0;
		nn = 0;
		if (token(cp+1,p-cp-1, // pos,cp-pos,
			  FI_POIN,  " ("," )[",pp,nn) != FI_POIN) {
if (!locale) {
		  c.err(cp+1,p-cp-1,"Must be a pointer.\n");
} else {
		  c.err(cp+1,p-cp-1,"  .\n");
}
                  c.flsh(); 
		  error_code = 5;
		  return error_code;
		}
		r_action.push_back(FI_POIN);
		poin_num.push_back(nn);
// find square brackets...
		b_a = pp;
		nb  = 0; // the number of dereferencings
		while (b_a != p) {
		  if (*b_a == '[') {
		    ++nb;
		    ++b_a;
		    char* bb_a = b_a;
		    pair(bb_a,p-bb_a,'[',']');
		    if (decompose(b_a, bb_a, 0, ccoun))
		      return error_code;
		    b_a = bb_a + 1;
		  } else if (*b_a != ' ') {
if (!locale) {
		    c.err(b_a,1,"Wrong expression.\n");
} else {
		    c.err(b_a,1," .\n");
}
                    c.flsh(); 
		    error_code = 13;
		    return error_code;
		  }
		}
		s_action.push_back(FI_POIN);
		poin_num_level.push_back(nb);
		if (usr_function_2p_poin_type.getPtr()[numb*2+1].level >
		    poin_type.getPtr()[nn].level +
		    poin_type.getPtr()[nn].num_dims - nb) {
if (!locale) {
		  c.err(cp+1,p-(cp+1), "Too many \'[]\'s.\n");
} else {
		  c.err(cp+1,p-(cp+1), "  \'[]\'.\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
		if (usr_function_2p_poin_type.getPtr()[numb*2+1].level <
		    poin_type.getPtr()[nn].level +
		    poin_type.getPtr()[nn].num_dims - nb) {
if (!locale) {
		  c.err(cp+1,p-(cp+1), "Too small \'[]\'s.\n");
} else {
		  c.err(cp+1,p-(cp+1), "  \'[]\'.\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
		if (poin_type.getPtr()[nn].num_dims < nb) {
if (!locale) {
		  c.err(cp+1,p-(cp+1),
			"Unknown dimension size(s).\n");
} else {
		  c.err(cp+1,p-(cp+1),
			"  .\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
// check second argument basic type...
		if (usr_function_2p_poin_type.getPtr()[numb*2+1].b_type !=
		    FI_USR_CLASS &&
		    usr_function_2p_poin_type.getPtr()[numb*2+1].b_type !=
		    poin_type.getPtr()[nn].b_type ||
		    usr_function_2p_poin_type.getPtr()[numb*2+1].b_type ==
		    FI_USR_CLASS &&
		    usr_function_2p_poin_type.getPtr()[numb*2+1].poin_type !=
		    poin_type.getPtr()[nn].poin_type) {
if (!locale) {
		  c.err(cp+1,p-(cp+1),
			"Formal and actual argument types mismatch (\'");
} else {
		  c.err(cp+1,p-(cp+1),
			"       (\'");
}
		  if (usr_function_2p_poin_type.getPtr()[numb*2+1].b_type !=
		      FI_USR_CLASS)
		    c.err(str_type[usr_function_2p_poin_type.
				  getPtr()[numb].b_type],5);
		  else
		    c.err(usr_function_2p_poin_type.getPtr()
			  [numb*2+1].poin_type,5);
if (!locale) {
		  c.err("\' and \'");
} else {
		  c.err("\'  \'");
}
		  if (poin_type.getPtr()[nn].b_type != FI_USR_CLASS)
		    c.err(str_type[poin_type.getPtr()[nn].b_type],5);
		  else
		    c.err(poin_type.getPtr()[nn].poin_type,5);
		  c.err("\').\n");
                  c.flsh(); 
		  error_code = 12;
		  return error_code;
		}
		usr_function_2p_num.push_back(numb);
	      } else if (command == FI_FUNC2) {
		//if (decompose(cp+1,p, 0))
		//  return error_code;
		if (decompose(pos,cp, 0, ccoun))
		  return error_code;
// find a generic pointer...
		char* pp = 0;
		uint  nn = 0;
		if (token(cp+1,p-(cp+1),
			  FI_POIN,  " ("," )[",pp,nn) != FI_POIN) {
if (!locale) {
		  c.err(cp+1,p-(cp+1),"Must be a pointer.\n");
} else {
		  c.err(cp+1,p-(cp+1),"  .\n");
}
                  c.flsh(); 
		  error_code = 5;
		  return error_code;
		}
		r_action.push_back(FI_POIN);
		poin_num.push_back(nn);
// find square brackets...
		char* b_a = pp;
		uint  nb  = 0; // the number of dereferencings
		while (b_a != p) {
		  if (*b_a == '[') {
		    ++nb;
		    ++b_a;
		    char* bb_a = b_a;
		    pair(bb_a,cp-bb_a,'[',']');
		    if (decompose(b_a, bb_a, 0, ccoun))
		      return error_code;
		    b_a = bb_a + 1;
		  } else if (*b_a != ' ') {
if (!locale) {
		    c.err(b_a,1,"Wrong expression.\n");
} else {
		    c.err(b_a,1," .\n");
}
                    c.flsh(); 
		    error_code = 13;
		    return error_code;
		  }
		}
		s_action.push_back(FI_POIN);
		poin_num_level.push_back(nb);
// check second argument pointer level,
// 1. for example, name "arr" of an array double[][][] is considered in a
//    parsing as double*** (of course, realy for auto and static memory this
//    is different type) and if the first argument in signature of the
//    function must have a type double* only two times dereferenced array name
//    (arr[][]) is right, but if array elements have a type double* all three
//    dereferencings must be.
// 2. of course, an array elements must be placed consequently (if one is
//    going to make dereferencings in formulas, because dereferencings are
//    made just by offsets of a pointer).
// such a way is used due to a condition of compatibility of names of auto
// and static arrays and dynamic memory pointers to n-dimentional arrays,
// one has to realize the points else very strange errors due to memory
// corruptions can happen.
		if (usr_function_2_poin_type.getPtr()[numb].level >
		    poin_type.getPtr()[nn].level +
		    poin_type.getPtr()[nn].num_dims - nb) {
if (!locale) {
		  c.err(cp+1,p-(cp+1), "Too many \'[]\'s.\n");
} else {
		  c.err(cp+1,p-(cp+1), "  \'[]\'.\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
		if (usr_function_2_poin_type.getPtr()[numb].level <
		    poin_type.getPtr()[nn].level +
		    poin_type.getPtr()[nn].num_dims - nb) {
if (!locale) {
		  c.err(cp+1,p-(cp+1), "Too small \'[]\'s.\n");
} else {
		  c.err(cp+1,p-(cp+1), "  \'[]\'.\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
		if (poin_type.getPtr()[nn].num_dims < nb) {
if (!locale) {
		  c.err(cp+1,p-(cp+1),
			"Unknown dimension size(s).\n");
} else {
		  c.err(cp+1,p-(cp+1),
			"  .\n");
}
                  c.flsh(); 
		  error_code = 11;
		  return error_code;
		}
// check second argument basic type...
		if (usr_function_2_poin_type.getPtr()[numb].b_type !=
		    FI_USR_CLASS &&
		    usr_function_2_poin_type.getPtr()[numb].b_type !=
		    poin_type.getPtr()[nn].b_type ||
		    usr_function_2_poin_type.getPtr()[numb].b_type ==
		    FI_USR_CLASS &&
		    usr_function_2_poin_type.getPtr()[numb].poin_type !=
		    poin_type.getPtr()[nn].poin_type) {
if (!locale) {
		  c.err(cp+1,p-(cp+1),
			"Formal and actual argument types mismatch (\'");
} else {
		  c.err(cp+1,p-(cp+1),
			"       (\'");
}
		  if (usr_function_2_poin_type.getPtr()[numb].b_type !=
		      FI_USR_CLASS)
		    c.err(str_type[usr_function_2_poin_type.
				  getPtr()[numb].b_type],5);
		  else
		    c.err(usr_function_2_poin_type.getPtr()[numb].poin_type,5);
if (!locale) {
		  c.err("\' and \'");
} else {
		  c.err("\'  \'");
}
		  if (poin_type.getPtr()[nn].b_type != FI_USR_CLASS)
		    c.err(str_type[poin_type.getPtr()[nn].b_type],5);
		  else
		    c.err(poin_type.getPtr()[nn].poin_type,5);
		  c.err("\').\n");
                  c.flsh(); 
		  error_code = 12;
		  return error_code;
		}
		usr_function_2_num.push_back(numb);
	      } else if (!cp) {
// unary function...
		if (decompose(pos, p, 0, ccoun))
		  return error_code;
	      } else {
// binary function...
		if (decompose(pos,cp, 0, ccoun))
		  return error_code;
		if (decompose(cp+1,p, 0, ccoun))
		  return error_code;
		if (command == FI_FUNC2P)
		  usr_function_2p_num.push_back(numb);
		else if (command == FI_FUNC2D)
		  usr_function_2d_num.push_back(numb);
	      }
	    }
	  } else {
if (!locale) {
	    c.err(expr,size,"After a function name the "
		  "round left bracket is waited.\n");
} else {
	    c.err(expr,size,
		  "      .\n");
}
            c.flsh(); 
	    error_code = 1;
	    return error_code;
	  }
	}
      }
      if (tail && tail != end) {
	uint rest = end-tail;
	while (rest--)
	  if (*tail != ' ') {
	    if (*tail == ')')
if (!locale) {
	      c.err(tail,1,
		    "Mismatched right round bracket.\n");
} else {
	      c.err(tail,1,
		    "    .\n");
}
	    else if (*tail == ']')
if (!locale) {
	      c.err(tail,1,
		    "Mismatched right square bracket.\n");
} else {
	      c.err(tail,1,
		    "    .\n");
}
	    else
if (!locale) {
	      c.err(tail,rest+1,
		    "Wrong token or wrong expression.\n");
} else {
	      c.err(tail,rest+1,
		    "    .\n");
}
            c.flsh(); 
	    error_code = 1;
	    return error_code;
	    ++tail;
	  }
      }
      if (command != FI_L_PARENTHESIS && command != FI_UN_PLUS)
	s_action.push_back(command);
      if (command >= FI_FUNC && command != FI_L_PARENTHESIS) ++ccoun;
    }
    break;

  default:;
  }
  return error_code;
}

void forint::
push_cast_pos(char* b, char* e)
{
  b_cast_pos.push_back(b-b_decomposed_string);
  e_cast_pos.push_back(e-b_decomposed_string);
}

void forint::
d_push_cast_pos(char* b, char* e)
{
  d_b_cast_pos.push_back(b-b_decomposed_string);
  d_e_cast_pos.push_back(e-b_decomposed_string);
}

void forint::
push_long_name_pos(char* b,string name)
{
  new_name_pos[b-b_decomposed_string] = name;
}

}
